#define _XOPEN_SOURCE
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <time.h>
#include <math.h>

#include <mpi.h>

#define max(a,b) ({ __typeof__ (a) _a = (a); __typeof__ (b) _b = (b); _a > _b ? _a : _b; })
#define min(a,b) ({ __typeof__ (a) _a = (a); __typeof__ (b) _b = (b); _a < _b ? _a : _b; })

size_t offset(size_t total, size_t rank, size_t size)
{
	return rank * (total / size) + min(total % size, rank);
}
size_t length(size_t total, size_t rank, size_t size)
{
	return total / size + (total % size > rank);
}

/*****************************************************************************
 *                       Matrice structure definition                        *
 *****************************************************************************/

typedef struct
{
	size_t rows;  // le nombre de lignes
	size_t cols;  // le nombre de colonnes
	double *data; // les valeurs
} matrix_t;

/*****************************************************************************
 *                          Initialize empty matrix                          *
 *****************************************************************************/
void matrix_empty(matrix_t *mat, size_t height, size_t width)
{
	mat->rows = height;
	mat->cols = width;
	mat->data = (double*) calloc(width*height, sizeof(double));
}

/*****************************************************************************
 *                         Initialize random matrix                          *
 *****************************************************************************/
void matrix_random(matrix_t *mat, size_t height, size_t width)
{
	matrix_empty(mat, height, width);
	// for (size_t i=0; i<width*height; ++i) mat->data[i] = drand48();
	for (size_t i=0; i<width*height; ++i) mat->data[i] = 1.0;
}

/*****************************************************************************
 *                           Delete matrix content                           *
 *****************************************************************************/
void matrix_delete(matrix_t *mat)
{
	mat->cols = 0;
	mat->rows = 0;
	free(mat->data);
}

/*****************************************************************************
 *                               Dispay matrix                               *
 *****************************************************************************/
void matrix_display(matrix_t A)
{
	if (A.rows && A.cols)
	{
		printf("       ");
		for(size_t j=0; j<A.cols; ++j) printf("    %.4lu   ",j);
		printf("\n     __");
		for(size_t j=0; j<A.cols; ++j) printf("___________");
		printf("_\n");
		for(size_t i=0; i<A.rows; ++i)
		{
			printf("%.4lu | ", i);
			for(size_t j=0; j<A.cols; ++j) printf("% 10.4f ",A.data[i*A.cols+j]);
			printf("|\n");
		}
		printf("     --");
		for(size_t j=0; j<A.cols; ++j) printf("-----------");
		printf("-\n");
	}
	else
	{
		printf("Empty matrix\n");
	}
}

/*****************************************************************************
 *                              Matrix mutliply                              *
 *****************************************************************************/
void matrix_multiply(matrix_t A, matrix_t B, matrix_t C)
{
	if ((A.rows != C.rows) || (A.cols != B.rows) || (B.cols != C.cols))
	{
		fprintf(
			stderr,
			"[ERROR] Incompatible matrix dimensions\n A(%lu,%lu) B(%lu,%lu) C(%lu,%lu)\n",
			A.rows, A.cols,
			B.rows, B.cols,
			C.rows, C.cols
		);
		exit(1);
	}
	size_t M = A.rows;
	size_t K = A.cols;
	size_t N = B.cols;

	// Can be used to reset C (not to do for iteratives accumulations)
	// memset(C.data, 0, C.cols * C.rows * sizeof(double));

	for (size_t i=0; i<M; ++i)
	for (size_t j=0; j<N; ++j)
	for (size_t k=0; k<K; ++k)
		C.data[i*N+j] += A.data[i*K+k] * B.data[k*N+j];
}

int main(int argc, char * argv[])
{

	size_t M = 13;
	size_t K = 17;
	size_t N = 19;



	MPI_Init(&argc, &argv);

	int rank, size;
	MPI_Comm_rank(MPI_COMM_WORLD, &rank);
	MPI_Comm_size(MPI_COMM_WORLD, &size);

	if (argc == 4)
	{
		M = atol(argv[1]);
		K = atol(argv[2]);
		N = atol(argv[3]);
	}
	else if (!rank)
	{
		fprintf(stderr, "Matrix dimensions can be provided as arguments <M> <K> <N>.\nUsing defaut values: M = %lu, K = %lu, N = %lu\n", M, K, N);
	}



	int p  = sqrt(size);
	int in = rank < p * p;

	MPI_Comm grid_comm;
	MPI_Comm_split(MPI_COMM_WORLD, in, rank, &grid_comm);


	if (in) // We only keep p*p procs alive (this in the grid)
	{
		srand48(time(NULL) + rank);

		// Initializa communicators
		MPI_Comm h_comm, v_comm;
		MPI_Comm_split(grid_comm, rank / p, rank, &h_comm);
		MPI_Comm_split(grid_comm, rank % p, rank, &v_comm);

		// get rank and size
		int hrank, hsize;
		int vrank, vsize;
		MPI_Comm_rank(h_comm, &hrank);
		MPI_Comm_size(h_comm, &hsize);
		MPI_Comm_rank(v_comm, &vrank);
		MPI_Comm_size(v_comm, &vsize);

		// Initialize matrices
		matrix_t A, B, C, tA, tB;
		matrix_random(&A, length(M, vrank, vsize), length(K, hrank, hsize));
		matrix_random(&B, length(K, vrank, vsize), length(N, hrank, hsize));
		matrix_empty (&C, length(M, vrank, vsize), length(N, hrank, hsize));

		for (int step=0; step<p; ++step)
		{
			// If we receive a block (rank != step) : alloc + bcast recv
			if (step != hrank)
			{
				matrix_empty(&tA, length(M, vrank, vsize), length(K, step, hsize));
				MPI_Bcast(tA.data, tA.rows*tA.cols, MPI_DOUBLE, step, h_comm);
			}
			// If we send a block (rank == step) : bcast send
			else
				MPI_Bcast(A.data, A.rows*A.cols, MPI_DOUBLE, step, h_comm);

			// If we receive a block (rank != step) : alloc + bcast recv
			if (step != vrank)
			{
				matrix_empty(&tB, length(K, step, vsize), length(N, hrank, hsize));
				MPI_Bcast(tB.data, tB.rows*tB.cols, MPI_DOUBLE, step, v_comm);
			}
			// If we send a block (rank == step) : bcast send
			else
				MPI_Bcast(B.data, B.rows*B.cols, MPI_DOUBLE, step, v_comm);

			// We multiply the blocks
			matrix_multiply((step==hrank)?A:tA, (step==vrank)?B:tB, C);

			// Memory cleanup
			if (step != hrank) matrix_delete(&tA);
			if (step != vrank) matrix_delete(&tB);
		}

		// Display distributed matrix
		for (int y=0; y<p; ++y)
		for (int x=0; x<p; ++x)
		{
			if (x == hrank && y == vrank)
			{
				printf("================= [%d] %dx%d =================\n", rank, x, y);
				matrix_display(C);
				printf("\n");
			}
			MPI_Barrier(grid_comm);
		}

		// Delete Matrices
		matrix_delete(&A);
		matrix_delete(&B);
		matrix_delete(&C);
	}

	MPI_Finalize();

	return 0;
}
